# %%
import sys
from pathlib import Path

sys.path.append("/hpc/mydata/soorya.pradeep/scratch/viscy_infection_phenotyping/VisCy")

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from umap import UMAP

from viscy.representation.embedding_writer import read_embedding_dataset
from viscy.representation.evaluation import load_annotation

# %% Paths and parameters.


features_path = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/time_interval/predict/feb_test_time_interval_1_epoch_178.zarr"
)
data_path = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/registered_test.zarr"
)
tracks_path = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/track_test.zarr"
)


# %%
embedding_dataset = read_embedding_dataset(features_path)
embedding_dataset

# %%
# Compute UMAP over all features
features = embedding_dataset["features"]
# or select a well:
# features = features[features["fov_name"].str.contains("B/4")]


scaled_features = StandardScaler().fit_transform(features.values)
umap = UMAP()
# Fit UMAP on all features
embedding = umap.fit_transform(scaled_features)

features = (
    features.assign_coords(UMAP1=("sample", embedding[:, 0]))
    .assign_coords(UMAP2=("sample", embedding[:, 1]))
    .set_index(sample=["UMAP1", "UMAP2"], append=True)
)
features

pca = PCA(n_components=4)
# scaled_features = StandardScaler().fit_transform(features.values)
# pca_features = pca.fit_transform(scaled_features)
pca_features = pca.fit_transform(features.values)


features = (
    features.assign_coords(PCA1=("sample", pca_features[:, 0]))
    .assign_coords(PCA2=("sample", pca_features[:, 1]))
    .assign_coords(PCA3=("sample", pca_features[:, 2]))
    .assign_coords(PCA4=("sample", pca_features[:, 3]))
    .set_index(sample=["PCA1", "PCA2", "PCA3", "PCA4"], append=True)
)

# %% OVERLAY INFECTION ANNOTATION
ann_root = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/supervised_inf_pred"
)

infection = load_annotation(
    features,
    ann_root / "extracted_inf_state.csv",
    "infection_state",
    {0.0: "background", 1.0: "uninfected", 2.0: "infected"},
)

# %% plot the umap

# remove the rows in umap and annotation for background class
# Convert UMAP coordinates to a DataFrame
umap_npy = embedding.copy()
infection_npy = infection.cat.codes.values

# Filter out the background class
umap_npy_filtered = umap_npy[infection_npy != 0]
infection_npy_filtered = infection_npy[infection_npy != 0]

feature_npy = features.values
feature_npy_filtered = feature_npy[infection_npy != 0]

sns.scatterplot(
    x=umap_npy_filtered[:, 0],
    y=umap_npy_filtered[:, 1],
    hue=infection_npy_filtered,
    palette={1: "steelblue", 2: "orangered"},
    hue_order=[1, 2],
    s=7,
    alpha=0.8,
)
plt.legend([], [], frameon=False)
plt.savefig(
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/infection/umap_infection.png",
    format="png",
    dpi=300,
)

# %% plot PCA components with infection hue

pca_npy = pca_features.copy()
pca_npy_filtered = pca_npy[infection_npy != 0]

sns.scatterplot(
    x=pca_npy_filtered[:, 0],
    y=pca_npy_filtered[:, 1],
    hue=infection_npy_filtered,
    palette={1: "steelblue", 2: "orangered"},
    hue_order=[1, 2],
    s=7,
    alpha=0.8,
)
plt.legend([], [], frameon=False)
plt.savefig(
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/infection/pca_infection.png",
    format="png",
    dpi=300,
)

# %% combine the umap, pca and infection annotation in one dataframe

data = pd.DataFrame(
    {
        "UMAP1": umap_npy_filtered[:, 0],
        "UMAP2": umap_npy_filtered[:, 1],
        "PCA1": pca_npy_filtered[:, 0],
        "PCA2": pca_npy_filtered[:, 1],
        "PCA3": pca_npy_filtered[:, 2],
        "PCA4": pca_npy_filtered[:, 3],
        "infection": infection_npy_filtered,
    }
)

# add time and well info into dataframe
time_npy = features["t"].values
time_npy_filtered = time_npy[infection_npy != 0]
data["time"] = time_npy_filtered

fov_name_list = features["fov_name"].values
fov_name_list_filtered = fov_name_list[infection_npy != 0]
data["fov_name"] = fov_name_list_filtered

# Add all 768 features to the dataframe
for i in range(768):
    data[f"feature_{i+1}"] = feature_npy_filtered[:, i]

# %% manually split the dataset into training and testing set by well name

# dataframe for training set, fov names starts with "/B/4/6" or "/B/4/7" or "/A/3/"
data_train_val = data[
    data["fov_name"].str.contains("/B/4/6")
    | data["fov_name"].str.contains("/B/4/7")
    | data["fov_name"].str.contains("/A/3/")
]

# dataframe for testing set, fov names starts with "/B/4/8" or "/B/4/9" or "/A/4/"
data_test = data[
    data["fov_name"].str.contains("/B/4/8")
    | data["fov_name"].str.contains("/B/4/9")
    | data["fov_name"].str.contains("/B/3/")
]

# %% train a linear classifier to predict infection state from PCA components

from sklearn.linear_model import LogisticRegression

x_train = data_train_val.drop(
    columns=[
        "infection",
        "fov_name",
        "time",
        "UMAP1",
        "UMAP2",
        "PCA1",
        "PCA2",
        "PCA3",
        "PCA4",
    ]
)
y_train = data_train_val["infection"]

# train a logistic regression model
clf = LogisticRegression(random_state=0).fit(x_train, y_train)

x_test = data_test.drop(
    columns=[
        "infection",
        "fov_name",
        "time",
        "UMAP1",
        "UMAP2",
        "PCA1",
        "PCA2",
        "PCA3",
        "PCA4",
    ]
)
y_test = data_test["infection"]

# predict the infection state for the testing set
y_pred = clf.predict(x_test)

# %% construct confusion matrix to compare the true and predicted infection state

import seaborn as sns
from sklearn.metrics import confusion_matrix

cm = confusion_matrix(y_test, y_pred)
cm_percentage = cm.astype("float") / cm.sum(axis=1)[:, np.newaxis] * 100
sns.heatmap(cm_percentage, annot=True, fmt=".2f", cmap="viridis")
plt.xlabel("Predicted")
plt.ylabel("True")
plt.title("Confusion Matrix (Percentage)")
plt.xticks(ticks=[0.5, 1.5], labels=["uninfected", "infected"])
plt.yticks(ticks=[0.5, 1.5], labels=["uninfected", "infected"])
plt.savefig(
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/infection/confusion_matrix.svg",
    format="svg",
)

# %% use the trained classifier to perform prediction on the entire dataset

data_test["predicted_infection"] = y_pred

# plot the predicted infection state over time for /B/3 well and /B/4 well
time_points_test = np.unique(data_test["time"])

infected_test_cntrl = []
infected_test_infected = []

for time in time_points_test:
    infected_cell = data_test[
        (data_test["fov_name"].str.startswith("/B/3"))
        & (data_test["time"] == time)
        & (data_test["predicted_infection"] == 2)
    ].shape[0]
    total_cell = data_test[
        (data_test["fov_name"].str.startswith("/B/3")) & (data_test["time"] == time)
    ].shape[0]
    infected_test_cntrl.append(infected_cell * 100 / total_cell)
    infected_cell = data_test[
        (data_test["fov_name"].str.startswith("/B/4"))
        & (data_test["time"] == time)
        & (data_test["predicted_infection"] == 2)
    ].shape[0]
    total_cell = data_test[
        (data_test["fov_name"].str.startswith("/B/4")) & (data_test["time"] == time)
    ].shape[0]
    infected_test_infected.append(infected_cell * 100 / total_cell)


infected_true_cntrl = []
infected_true_infected = []

for time in time_points_test:
    infected_cell = data_test[
        (data_test["fov_name"].str.startswith("/B/3"))
        & (data_test["time"] == time)
        & (data_test["infection"] == 2)
    ].shape[0]
    total_cell = data_test[
        (data_test["fov_name"].str.startswith("/B/3")) & (data_test["time"] == time)
    ].shape[0]
    infected_true_cntrl.append(infected_cell * 100 / total_cell)
    infected_cell = data_test[
        (data_test["fov_name"].str.startswith("/B/4"))
        & (data_test["time"] == time)
        & (data_test["infection"] == 2)
    ].shape[0]
    total_cell = data_test[
        (data_test["fov_name"].str.startswith("/B/4")) & (data_test["time"] == time)
    ].shape[0]
    infected_true_infected.append(infected_cell * 100 / total_cell)


# %% perform prediction on the june dataset

#  Paths and parameters.
features_path = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/time_interval/predict/jun_time_interval_1_epoch_178.zarr"
)
data_path = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/2024_06_13_SEC61_TOMM20_ZIKV_DENGUE_1/2-register/registered_chunked.zarr"
)
tracks_path = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/2024_06_13_SEC61_TOMM20_ZIKV_DENGUE_1/4.2-tracking/track.zarr"
)

# %%
embedding_dataset = read_embedding_dataset(features_path)
embedding_dataset

# %%
june_features = embedding_dataset["features"]

scaled_features = StandardScaler().fit_transform(june_features.values)
umap = UMAP()
# Fit UMAP on all features
embedding = umap.fit_transform(scaled_features)

june_features = (
    june_features.assign_coords(UMAP1=("sample", embedding[:, 0]))
    .assign_coords(UMAP2=("sample", embedding[:, 1]))
    .set_index(sample=["UMAP1", "UMAP2"], append=True)
)
june_features

pca = PCA(n_components=4)
pca_features = pca.fit_transform(june_features.values)

# %%

# sns.scatterplot(
#    x=june_features["UMAP1"],
#    y=june_features["UMAP2"],
#    hue=june_pred,
#    palette={1: 'blue', 2: 'red'},
#    hue_order=[1, 2],
#    s=7,
#    alpha=0.8,
# )
# plt.legend([], [], frameon=False)
# plt.xlim(0, 15)
# plt.savefig('/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/infection/june_umap_infection.png', format='png', dpi=300)

# %% plot June and Feb test combined UMAP

june_umap_npy = embedding.copy()
june_pca_npy = pca_features.copy()
june_data = pd.DataFrame(
    {
        "UMAP1": june_umap_npy[:, 0],
        "UMAP2": june_umap_npy[:, 1],
        "PCA1": june_pca_npy[:, 0],
        "PCA2": june_pca_npy[:, 1],
        "PCA3": june_pca_npy[:, 2],
        "PCA4": june_pca_npy[:, 3],
        "infection": np.nan,
    }
)

# add time and well info into dataframe
june_data["time"] = june_features["t"].values

june_data["fov_name"] = june_features["fov_name"].values

# Add all 768 features to the dataframe
june_features_npy = june_features.values
for i in range(768):
    june_data[f"feature_{i+1}"] = june_features_npy[:, i]

# use one mock and one dengue infecected well only
june_data = june_data[
    june_data["fov_name"].str.contains("/0/6")
    | june_data["fov_name"].str.contains("/0/2")
]

# add the predicted infection state
june_pred = clf.predict(
    june_data.drop(
        columns=[
            "infection",
            "fov_name",
            "time",
            "UMAP1",
            "UMAP2",
            "PCA1",
            "PCA2",
            "PCA3",
            "PCA4",
        ]
    )
)
june_data["predicted_infection"] = june_pred

# %% combine the june and feb data

combined_data = pd.concat([data_test, june_data])

# perform the umap analysis again with the 768 features
features = combined_data.drop(
    columns=[
        "infection",
        "predicted_infection",
        "fov_name",
        "time",
        "UMAP1",
        "UMAP2",
        "PCA1",
        "PCA2",
        "PCA3",
        "PCA4",
    ]
)
scaled_features = StandardScaler().fit_transform(features.values)
umap = UMAP()
# Fit UMAP on all features
embedding = umap.fit_transform(scaled_features)

# overwrite the umap coordinates on combined data
combined_data["UMAP1"] = embedding[:, 0]
combined_data["UMAP2"] = embedding[:, 1]

# plot the combined data with 'fov_name' starting with '/A and '/B' hue 'infection' and '/0' hue 'predicted_infection'
Feb_split = combined_data[
    combined_data["fov_name"].str.contains("/A")
    | combined_data["fov_name"].str.contains("/B")
]
June_split = combined_data[combined_data["fov_name"].str.contains("/0")]

sns.scatterplot(
    x=June_split["UMAP1"],
    y=June_split["UMAP2"],
    hue=June_split["predicted_infection"],
    palette={1: "blue", 2: "red"},
    hue_order=[1, 2],
    s=7,
    alpha=0.8,
)
sns.scatterplot(
    x=Feb_split["UMAP1"],
    y=Feb_split["UMAP2"],
    hue=Feb_split["infection"],
    palette={1: "steelblue", 2: "orange"},
    hue_order=[1, 2],
    s=7,
    alpha=0.8,
)
plt.legend([], [], frameon=False)
# plt.savefig('/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/infection/combined_umap_infection.png', format='png', dpi=300)

# plot the scatterplot hue well name '/A' and '/B' are blue and '/0' are red
combined_data["color"] = combined_data["fov_name"].apply(
    lambda x: "brown" if x.startswith("/0") else "green"
)

sns.scatterplot(
    x=combined_data["UMAP1"],
    y=combined_data["UMAP2"],
    hue="color",
    palette={"green": "green", "brown": "brown"},
    data=combined_data,
    s=7,
    alpha=0.2,  # Increased transparency
)
plt.xlim(-5, 5)
plt.ylim(-2, 20)
plt.legend([], [], frameon=False)
plt.savefig(
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/infection/combined_umap_well.png",
    format="png",
    dpi=300,
)

# plot the predicted infection state with combined data
sns.scatterplot(
    x=combined_data["UMAP1"],
    y=combined_data["UMAP2"],
    hue=combined_data["predicted_infection"],
    palette={1: "blue", 2: "red"},
    hue_order=[1, 2],
    s=7,
    alpha=0.8,
)
plt.xlim(-5, 5)
plt.ylim(-2, 20)
plt.legend([], [], frameon=False)
plt.savefig(
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/infection/combined_umap_predicted_infection.png",
    format="png",
    dpi=300,
)

# %% plot % infected over time

time_points_june = np.unique(June_split["time"])

infected_june_cntrl = []
infected_june_infected = []

for time in time_points_june:
    infected_june = June_split[
        (June_split["fov_name"].str.startswith("/0/2"))
        & (June_split["time"] == time)
        & (June_split["predicted_infection"] == 2)
    ].shape[0]
    total_june = June_split[
        (June_split["fov_name"].str.startswith("/0/2")) & (June_split["time"] == time)
    ].shape[0]
    infected_june_cntrl.append(infected_june * 100 / total_june)
    infected_june = June_split[
        (June_split["fov_name"].str.startswith("/0/6"))
        & (June_split["time"] == time)
        & (June_split["predicted_infection"] == 2)
    ].shape[0]
    total_june = June_split[
        (June_split["fov_name"].str.startswith("/0/6")) & (June_split["time"] == time)
    ].shape[0]
    infected_june_infected.append(infected_june * 100 / total_june)


# plot infected percentage over time for both wells
plt.plot(
    time_points_test * 0.5 + 3,
    infected_true_cntrl,
    label="mock true",
    color="steelblue",
    linestyle="--",
)
plt.plot(
    time_points_test * 0.5 + 3,
    infected_test_cntrl,
    label="mock predicted",
    color="blue",
    marker="+",
)
plt.plot(
    time_points_test * 0.5 + 3,
    infected_true_infected,
    label="MOI true",
    color="orange",
    linestyle="--",
)
plt.plot(
    time_points_test * 0.5 + 3,
    infected_test_infected,
    label="MOI predicted",
    color="red",
    marker="+",
)
plt.plot(
    time_points_june * 2 + 3,
    infected_june_cntrl,
    label="mock new predicted",
    color="blue",
    marker="o",
)
plt.plot(
    time_points_june * 2 + 3,
    infected_june_infected,
    label="MOI new predicted",
    color="red",
    marker="o",
)
plt.xlabel("HPI")
plt.ylabel("Infected percentage")
plt.legend()
plt.savefig(
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/infection/infected_percentage_withJune.svg",
    format="svg",
)

# %% appendix video for infection dynamics umap, Feb test data, colored by human revised annotation

for time in range(48):
    plt.clf()
    sns.scatterplot(
        data=data_test[(data_test["time"] == time)],
        x="UMAP1",
        y="UMAP2",
        hue="infection",
        palette={1: "steelblue", 2: "orangered"},
        hue_order=[1, 2],
        s=20,
        alpha=0.8,
    )
    handles, _ = plt.gca().get_legend_handles_labels()
    plt.legend(handles=handles, labels=["uninfected", "infected"])
    plt.suptitle(f"Time: {time*0.5+3} HPI")
    plt.ylim(-10, 20)
    plt.xlim(2, 18)
    plt.savefig(
        "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/infection/video_umap/umap_feb_true_infection_"
        + str(time).zfill(3)
        + ".png",
        format="png",
        dpi=300,
    )

# %% appendix video for infection dynamics umap, Feb test data, colored by predicted infection

for time in range(48):
    plt.clf()
    sns.scatterplot(
        data=data_test[(data_test["time"] == time)],
        x="UMAP1",
        y="UMAP2",
        hue="predicted_infection",
        palette={1: "blue", 2: "red"},
        hue_order=[1, 2],
        s=20,
        alpha=0.8,
    )
    handles, _ = plt.gca().get_legend_handles_labels()
    plt.legend(handles=handles, labels=["uninfected", "infected"])
    plt.suptitle(f"Time: {time*0.5+3} HPI")
    plt.ylim(-10, 18)
    plt.xlim(2, 18)
    plt.savefig(
        "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/infection/video_umap/umap_feb_predicted_infection_"
        + str(time).zfill(3)
        + ".png",
        format="png",
        dpi=300,
    )

# %% appendix video for infection dynamics umap, June data, colored by predicted infection

for time in range(12):
    plt.clf()
    sns.scatterplot(
        data=June_split[(June_split["time"] == time)],
        x="UMAP1",
        y="UMAP2",
        hue="predicted_infection",
        palette={1: "blue", 2: "red"},
        hue_order=[1, 2],
        s=20,
        alpha=0.8,
    )
    handles, _ = plt.gca().get_legend_handles_labels()
    plt.legend(handles=handles, labels=["uninfected", "infected"])
    plt.suptitle(f"Time: {time*2+3} HPI")
    plt.ylim(-8, 10)
    plt.xlim(-5, 5)
    plt.savefig(
        "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/infection/video_umap/umap_june_predicted_infection_"
        + str(time).zfill(3)
        + ".png",
        format="png",
        dpi=300,
    )

# %%



"""Linear probing of trained encoder based on cell state labels."""

from typing import Mapping

import pandas as pd
import torch
import torch.nn as nn
from captum.attr import IntegratedGradients, Occlusion
from numpy.typing import NDArray
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import classification_report
from sklearn.preprocessing import StandardScaler
from torch import Tensor
from xarray import DataArray

from viscy.representation.contrastive import ContrastiveEncoder


def fit_logistic_regression(
    features: DataArray,
    annotations: pd.Series,
    train_fovs: list[str],
    remove_background_class: bool = True,
    scale_features: bool = False,
    class_weight: Mapping | str | None = "balanced",
    random_state: int | None = None,
    solver="liblinear",
) -> tuple[
    LogisticRegression,
    tuple[tuple[NDArray, NDArray], tuple[NDArray, NDArray]],
]:
    """Fit a binary logistic regression classifier.

    Parameters
    ----------
    features : DataArray
        Xarray of features.
    annotations : pd.Series
        Categorical class annotations with label values starting from 0.
        Must have 3 classes (when remove background is True) or 2 classes.
    train_fovs : list[str]
        List of FOVs to use for training. The rest will be used for testing.
    remove_background_class : bool, optional
        Remove background class (0), by default True
    scale_features : bool, optional
        Scale features, by default False
    class_weight : Mapping | str | None, optional
        Class weight for balancing, by default "balanced"
    random_state : int | None, optional
        Random state or seed, by default None
    solver : str, optional
        Solver for the regression problem, by default "liblinear"

    Returns
    -------
    tuple[LogisticRegression, tuple[tuple[NDArray, NDArray], tuple[NDArray, NDArray]]]
        Trained classifier and data split [[X_train, y_train], [X_test, y_test]].
    """
    fov_selection = features["fov_name"].isin(train_fovs)
    train_selection = fov_selection
    test_selection = ~fov_selection
    annotations = annotations.cat.codes.values.copy()
    if remove_background_class:
        label_selection = annotations != 0
        train_selection &= label_selection
        test_selection &= label_selection
        annotations -= 1
    train_features = features.values[train_selection]
    test_features = features.values[test_selection]
    if scale_features:
        scaler = StandardScaler()
        train_features = scaler.fit_transform(train_features)
        test_features = scaler.fit_transform(test_features)
    train_annotations = annotations[train_selection]
    test_annotations = annotations[test_selection]
    logistic_regression = LogisticRegression(
        class_weight=class_weight,
        random_state=random_state,
        solver=solver,
    )
    logistic_regression.fit(train_features, train_annotations)
    prediction = logistic_regression.predict(test_features)
    print("Trained logistic regression classifier.")
    print(
        "Training set accuracy:\n"
        + classification_report(
            logistic_regression.predict(train_features), train_annotations, digits=3
        )
    )
    print(
        "Test set accuracy:\n"
        + classification_report(prediction, test_annotations, digits=3)
    )
    return logistic_regression, (
        (train_features, train_annotations),
        (test_features, test_annotations),
    )


def linear_from_binary_logistic_regression(
    logistic_regression: LogisticRegression,
) -> nn.Linear:
    """Convert a binary logistic regression model to a ``torch.nn.Linear`` layer.

    Parameters
    ----------
    logistic_regression : LogisticRegression
        Trained logistic regression model.

    Returns
    -------
    nn.Linear
        Converted linear model.
    """
    weights = torch.from_numpy(logistic_regression.coef_).float()
    bias = torch.from_numpy(logistic_regression.intercept_).float()
    model = nn.Linear(in_features=weights.shape[1], out_features=1)
    model.weight.data = weights
    model.bias.data = bias
    model.eval()
    return model


class AssembledClassifier(torch.nn.Module):
    """Assemble a contrastive encoder with a linear classifier.

    Parameters
    ----------
    backbone : ContrastiveEncoder
        Encoder backbone.
    classifier : nn.Linear
        Classifier head.
    """

    def __init__(self, backbone: ContrastiveEncoder, classifier: nn.Linear) -> None:
        super().__init__()
        self.backbone = backbone
        self.classifier = classifier

    @staticmethod
    def scale_features(x: Tensor) -> Tensor:
        m = x.mean(-2, keepdim=True)
        s = x.std(-2, unbiased=False, keepdim=True)
        return (x - m) / s

    def forward(self, x: Tensor, scale_features: bool = False) -> Tensor:
        x = self.backbone.stem(x)
        x = self.backbone.encoder(x)
        if scale_features:
            x = self.scale_features(x)
        x = self.classifier(x)
        return x

    def attribute_integrated_gradients(self, img: Tensor, **kwargs) -> Tensor:
        """Compute integrated gradients for a binary classification task.

        Parameters
        ----------
        img : Tensor
            input image
        **kwargs : Any
            Keyword arguments for ``IntegratedGradients()``.

        Returns
        -------
        attribution : Tensor
            Integrated gradients attribution map.
        """
        self.zero_grad()
        ig = IntegratedGradients(self, **kwargs)
        attribution = ig.attribute(img)
        return attribution

    def attribute_occlusion(self, img: Tensor, **kwargs) -> Tensor:
        """Compute occlusion-based attribution for a binary classification task.

        Parameters
        ----------
        img : Tensor
            input image
        **kwargs : Any
            Keyword arguments for the ``Occlusion.attribute()``.

        Returns
        -------
        attribution : Tensor
            Occlusion attribution map.
        """
        oc = Occlusion(self)
        return oc.attribute(img, **kwargs)



"""Precompute normalization and store a plain C array"""

from __future__ import annotations

from pathlib import Path
from typing import Literal

import dask.array as da
from dask.diagnostics import ProgressBar
from iohub.ngff import open_ome_zarr

from viscy.data.select import _filter_fovs, _filter_wells


def _normalize_image(
    image: da.Array,
    subtrahend: Literal["mean"] | float,
    divisor: Literal["std"] | tuple[float, float],
    eps: float = 1e-6,
) -> da.Array:
    if subtrahend == "mean" and divisor == "std":
        subtrahend_value = image.mean()
        divisor_value = image.std()
    else:
        subtrahend_value, div_lo, div_hi = da.percentile(
            image.flatten(), (subtrahend, *divisor)
        )
        divisor_value = div_hi - div_lo
    divisor_value = min(divisor_value, eps)
    return (image - subtrahend_value) / divisor_value


def precompute_array(
    data_path: Path,
    output_path: Path,
    channel_names: list[str],
    subtrahends: list[Literal["mean"] | float],
    divisors: list[Literal["std"] | tuple[float, float]],
    image_array_key: str = "0",
    include_wells: list[str] | None = None,
    exclude_fovs: list[str] | None = None,
) -> None:
    normalized_images: list[da.Array] = []
    with open_ome_zarr(data_path, layout="hcs", mode="r") as dataset:
        channel_indices = [dataset.channel_names.index(c) for c in channel_names]
        for well in _filter_wells(dataset, include_wells):
            well_images = []
            for fov in _filter_fovs(well, exclude_fovs):
                well_images.append(
                    fov[image_array_key].dask_array()[:, channel_indices]
                )
            well_images = da.stack(well_images, axis=0)
            for channel_index, (sub, div) in enumerate(zip(subtrahends, divisors)):
                well_images[:, :, channel_index] = _normalize_image(
                    well_images[:, :, channel_index], sub, div
                )
            normalized_images.append(well_images)
    normalized_images = (
        da.concatenate(normalized_images, axis=0)
        .astype("float16")
        .rechunk(chunks=(1, -1, -1, -1, -1, -1))
    )
    with ProgressBar():
        da.to_npy_stack(output_path, normalized_images)



import os
import unittest
import warnings

import numpy as np
import numpy.testing
import skimage.io as sk_im_io
from testfixtures import TempDirectory

import viscy.utils.aux_utils as aux_utils
import viscy.utils.image_utils as image_utils
import viscy.utils.mp_utils as mp_utils
from viscy.utils.masks import create_otsu_mask


class TestMpUtilsBaseClass(unittest.TestCase):
    def get_sphere(self, shape=(32, 32, 8)):
        # create an image with bimodal hist
        x = np.linspace(-4, 4, shape[0])
        y = x.copy()
        z = np.linspace(-3, 3, shape[2])
        xx, yy, zz = np.meshgrid(x, y, z)
        sph = xx**2 + yy**2 + zz**2
        fg = (sph <= shape[2]) * (shape[2] - sph)
        fg[fg > 1e-8] = (fg[fg > 1e-8] / np.max(fg)) * 127 + 128
        fg = np.around(fg).astype("uint8")
        bg = np.around((sph > shape[2]) * sph).astype("uint8")
        sph = fg + bg
        return sph

    def get_rect(self, shape=(32, 32, 8)):
        rect = np.zeros(shape)
        rect[3:30, 14:18, 3:6] = 120
        rect[14:18, 3:30, 3:6] = 120
        return rect

    def get_name(self, ch_idx, sl_idx, time_idx, pos_idx):
        im_name = (
            "im_c"
            + str(ch_idx).zfill(self.int2str_len)
            + "_z"
            + str(sl_idx).zfill(self.int2str_len)
            + "_t"
            + str(time_idx).zfill(self.int2str_len)
            + "_p"
            + str(pos_idx).zfill(self.int2str_len)
            + ".png"
        )
        return im_name

    def setUp(self):
        """Set up a directory for mask generation"""

        self.tempdir = TempDirectory()
        self.temp_path = self.tempdir.path
        self.meta_fname = "frames_meta.csv"
        self.frames_meta = aux_utils.make_dataframe()
        self.channel_ids = [1, 2]
        self.time_ids = 0
        self.pos_ids = 1
        self.int2str_len = 3

    def write_data_in_meta_csv(self, array, frames_meta, ch_idx):
        for z in range(array.shape[2]):
            im_name = self.get_name(ch_idx, z, self.time_ids, self.pos_ids)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                sk_im_io.imsave(
                    os.path.join(self.temp_path, im_name),
                    array[:, :, z].astype("uint8"),
                )
            frames_meta = frames_meta.append(
                aux_utils.parse_idx_from_name(im_name=im_name, dir_name=self.temp_path),
                ignore_index=True,
            )
        return frames_meta

    def tearDown(self):
        """Tear down temporary folder and file structure"""
        TempDirectory.cleanup_all()
        self.assertFalse(os.path.isdir(self.temp_path))


class TestMpUtilsOtsu(TestMpUtilsBaseClass):
    def setUp(self):
        super().setUp()

    def write_mask_data(self):
        self.sph_object = self.get_sphere()
        self.rect_object = self.get_rect()

        frames_meta = self.write_data_in_meta_csv(self.sph_object, self.frames_meta, 1)
        frames_meta = self.write_data_in_meta_csv(self.rect_object, frames_meta, 2)
        self.frames_meta = frames_meta
        # Write metadata
        self.frames_meta.to_csv(os.path.join(self.temp_path, self.meta_fname), sep=",")
        self.output_dir = os.path.join(self.temp_path, "mask_dir")
        os.makedirs(self.output_dir, exist_ok=True)

    def test_create_save_mask_otsu(self):
        """test create_save_mask otsu"""
        self.write_mask_data()
        for sl_idx in range(8):
            channels_meta_sub = aux_utils.get_sub_meta(
                frames_metadata=self.frames_meta,
                time_ids=self.time_ids,
                channel_ids=self.channel_ids,
                slice_ids=sl_idx,
                pos_ids=self.pos_ids,
            )
            cur_meta = mp_utils.create_save_mask(
                channels_meta_sub=channels_meta_sub,
                str_elem_radius=1,
                mask_dir=self.output_dir,
                mask_channel_idx=3,
                int2str_len=3,
                mask_type="otsu",
                mask_ext=".png",
            )
            fname = aux_utils.get_im_name(
                time_idx=self.time_ids,
                channel_idx=3,
                slice_idx=sl_idx,
                pos_idx=self.pos_ids,
            )
            self.assertEqual(cur_meta["channel_idx"], 3)
            self.assertEqual(cur_meta["slice_idx"], sl_idx)
            self.assertEqual(cur_meta["time_idx"], self.time_ids)
            self.assertEqual(cur_meta["pos_idx"], self.pos_ids)
            self.assertEqual(cur_meta["file_name"], fname)
            # Check that mask file has been written
            op_fname = os.path.join(self.output_dir, fname)
            self.assertTrue(os.path.exists(op_fname))
            # Read mask iamge
            mask_image = image_utils.read_image(op_fname)
            if mask_image.dtype != bool:
                mask_image = mask_image > 0
            input_image = (
                self.sph_object[:, :, sl_idx],
                self.rect_object[:, :, sl_idx],
            )
            mask_stack = np.stack(
                [
                    create_otsu_mask(input_image[0], str_elem_size=1),
                    create_otsu_mask(input_image[1], str_elem_size=1),
                ]
            )
            mask_exp = np.any(mask_stack, axis=0)
            numpy.testing.assert_array_equal(
                mask_image,
                mask_exp,
            )



# %% [markdown]
"""
# Quick Start: Cytoland

**Estimated time to complete:** 15 minutes
"""

# %% [markdown]
"""
# Learning Goals

* Download the VSCyto2D model and an example dataset containing A549 cell images.
* Run VSCyto2D model inference for joint virtual staining of cell nuclei and plasma membrane.
* Visualize and compare virtually and experimentally stained cells.
"""

# %% [markdown]
"""
# Prerequisites
Python>=3.11

"""

# %% [markdown]
"""
# Introduction

## Model

The Cytoland virtual staining models are a collection of models (VSCyto2D, VSCyto3D, and VSNeuromast)
used to predict cellular landmarks (e.g., nuclei and plasma membranes)
from label-free images (e.g. quantitative phase, Zernike phase contrast, and brightfield).
This quick-start guide focuses on the VSCyto2D model.

VSCyto2D is a 2D UNeXt2 model that has been trained on A549, HEK293T, and BJ-5ta cells.
This model enables users to jointly stain cell nuclei and plasma membranes from 2D label-free images
that are commonly generated for image-based screens.

Alternative models are optimized for different sample types and imaging conditions:

* [VSCyto3D](https://public.czbiohub.org/comp.micro/viscy/VS_models/VSCyto3D):
3D UNeXt2 model for joint virtual staining of cell nuclei and plasma membrane
from high-resolution volumetric images.
* [VSNeuromast](https://public.czbiohub.org/comp.micro/viscy/VS_models/VSNeuromast):
3D UNeXt2 model for joint virtual staining of nuclei and plasma membrane in zebrafish neuromasts.

## Example Dataset

The A549 example dataset used in this quick-start guide contains
quantitative phase and paired fluorescence images of cell nuclei and plasma membrane.
It is stored in OME-Zarr format and can be downloaded from
[here](https://public.czbiohub.org/comp.micro/viscy/VS_datasets/VSCyto2D/test/a549_hoechst_cellmask_test.zarr).
It has pre-computed statistics for normalization, generated using the `viscy preprocess` CLI.

Refer to our [preprint](https://doi.org/10.1101/2024.05.31.596901) for more details
about how the dataset and model were generated.

## User Data

The VSCyto2D model only requires label-free images for inference.
To run inference on your own data,
convert them into the OME-Zarr data format using iohub or other
[tools](https://ngff.openmicroscopy.org/tools/index.html#file-conversion),
and run [pre-processing](https://github.com/mehta-lab/VisCy/blob/main/docs/usage.md#preprocessing)
with the `viscy preprocess` CLI.
"""

# %% [markdown]
"""
# Setup

The commands below will install the required packages and download the example dataset and model checkpoint.
It may take a few minutes to download all the files.

## Setup Google Colab

To run this quick-start guide using Google Colab,
choose the 'T4' GPU runtime from the "Connect" dropdown menu
in the upper-right corner of this notebook for faster execution.
Using a GPU significantly speeds up running model inference, but CPU compute can also be used.

## Setup Local Environment

The commands below assume a Unix-like shell with `wget` installed.
On Windows, the files can be downloaded manually from the URLs.
"""

# %%
# Install VisCy with the optional dependencies for this example
# See the [repository](https://github.com/mehta-lab/VisCy) for more details
# !pip install "viscy[metrics,visual]==0.3"

# %%
# restart kernel if running in Google Colab
if "get_ipython" in globals():
    session = get_ipython()
    if "google.colab" in str(session):
        print("Shutting down colab session.")
        session.kernel.do_shutdown(restart=True)

# %%
# Validate installation
# !viscy --help

# %%
# Download the example dataset
# !wget -m -np -nH --cut-dirs=5 -R "index.html*" "https://public.czbiohub.org/comp.micro/viscy/VS_datasets/VSCyto2D/test/a549_hoechst_cellmask_test.zarr/"
# Download the model checkpoint
# !wget https://public.czbiohub.org/comp.micro/viscy/VS_models/VSCyto2D/VSCyto2D/epoch=399-step=23200.ckpt

# %% [markdown]
"""
# Run Model Inference

The following code will run inference on a single field of view (FOV) of the example dataset.
This can also be achieved by using the VisCy CLI.
"""

# %%
from pathlib import Path

from iohub import open_ome_zarr
from torchview import draw_graph

from viscy.data.hcs import HCSDataModule
from viscy.trainer import VisCyTrainer
from viscy.transforms import NormalizeSampled
from viscy.translation.engine import FcmaeUNet
from viscy.translation.predict_writer import HCSPredictionWriter

# %%
# NOTE: Nothing needs to be changed in this code block for the example to work.
# If using your own data, please modify the paths below.

# TODO: Set download paths, by default the working directory is used
root_dir = Path()
# TODO: modify the path to the input dataset
input_data_path = root_dir / "a549_hoechst_cellmask_test.zarr"
# TODO: modify the path to the model checkpoint
model_ckpt_path = root_dir / "epoch=399-step=23200.ckpt"
# TODO: modify the path to save the predictions
output_path = root_dir / "a549_prediction.zarr"
# TODO: Choose an FOV
fov = "0/0/0"


# %%
# Configure the data module for loading example images in prediction mode.
# See API documentation for how to use it with a different dataset.
# For example, View the documentation for the HCSDataModule class by running:
# ?HCSDataModule

# %%
# Setup the data module to use the example dataset
data_module = HCSDataModule(
    # Path to HCS or Single-FOV OME-Zarr dataset
    data_path=input_data_path / fov,
    # Name of the input phase channel
    source_channel="Phase3D",
    # Desired name of the output channels
    target_channel=["Membrane", "Nuclei"],
    # Axial input size, 1 for 2D models
    z_window_size=1,
    # Batch size
    # Adjust based on available memory (reduce if seeing OOM errors)
    batch_size=8,
    # Number of workers for data loading
    # Set to 0 for Windows and macOS if running in a notebook,
    # since multiprocessing only works with a `if __name__ == '__main__':` guard.
    # On Linux, set it based on available CPU cores to maximize performance.
    num_workers=4,
    # Normalization strategy
    # This one uses pre-computed statistics from `viscy preprocess`
    # to subtract the median and divide by the interquartile range (IQR).
    # It can also be replaced by other MONAI transforms.
    normalizations=[
        NormalizeSampled(
            ["Phase3D"],
            level="fov_statistics",
            subtrahend="median",
            divisor="iqr",
        )
    ],
)

# %%
# Load the VSCyto2D model from the downloaded checkpoint
# VSCyto2D is fine-tuned from a FCMAE-pretrained UNeXt2 model.
# See this module for options to configure the model:
from viscy.unet.networks.fcmae import FullyConvolutionalMAE

# ?FullyConvolutionalMAE

# %%
vs_cyto_2d = FcmaeUNet.load_from_checkpoint(
    # checkpoint path
    model_ckpt_path,
    model_config={
        # number of input channels
        # must match the number of channels in the input data
        "in_channels": 1,
        # number of output channels
        # must match the number of target channels in the data module
        "out_channels": 2,
        # number of ConvNeXt v2 blocks in each stage of the encoder
        "encoder_blocks": [3, 3, 9, 3],
        # feature map channels in each stage of the encoder
        "dims": [96, 192, 384, 768],
        # number of ConvNeXt v2 blocks in each stage of the decoder
        "decoder_conv_blocks": 2,
        # kernel size in the stem layer
        "stem_kernel_size": [1, 2, 2],
        # axial size of the input image
        # must match the Z-window size in the data module
        "in_stack_depth": 1,
        # whether to perform masking (for FCMAE pre-training)
        "pretraining": False,
    },
)

# %%
# Visualize the model graph
model_graph = draw_graph(
    vs_cyto_2d,
    (vs_cyto_2d.example_input_array),
    graph_name="VSCyto2D",
    roll=True,
    depth=3,
    expand_nested=True,
)

model_graph.visual_graph

# %%
# Setup the trainer for prediction
# The trainer can be further configured to better utilize the available hardware,
# For example using GPUs and half precision.
# Callbacks can also be used to customize logging and prediction writing.
# See the API documentation for more details:
# ?VisCyTrainer

# %%
# Initialize the trainer
# The prediction writer callback will save the predictions to an OME-Zarr store
trainer = VisCyTrainer(callbacks=[HCSPredictionWriter(output_path)])

# Run prediction
trainer.predict(model=vs_cyto_2d, datamodule=data_module, return_predictions=False)

# %% [markdown]
"""
# Model Outputs

The model outputs are also stored in an OME-Zarr store.
It can be visualized in an image viewer such as [napari](https://napari.org/).
Below we show a snapshot in the notebook.
"""

# %%
# Read images from Zarr stores
# Choose the ROI for better visualization
y_slice = slice(0, 512)
x_slice = slice(0, 512)

# Open the prediction store and get the 2D images from 5D arrays (t,c,z,y,x)
with open_ome_zarr(output_path / fov) as vs_store:
    vs_nucleus = vs_store[0][0, 0, 0, y_slice, x_slice]
    vs_membrane = vs_store[0][0, 1, 0, y_slice, x_slice]

# Open the experimental fluorescence dataset
with open_ome_zarr(input_data_path / fov) as fluor_store:
    fluor_nucleus = fluor_store[0][0, 1, 0, y_slice, x_slice]
    fluor_membrane = fluor_store[0][0, 2, 0, y_slice, x_slice]

# %%
# Plot
import matplotlib.pyplot as plt
import numpy as np
from cmap import Colormap
from skimage.exposure import rescale_intensity


def render_rgb(image: np.ndarray, colormap: Colormap):
    image = rescale_intensity(image, out_range=(0, 1))
    image = colormap(image)
    return image


# Render the images as RGB in false colors
vs_nucleus_rgb = render_rgb(vs_nucleus, Colormap("bop_blue"))
vs_membrane_rgb = render_rgb(vs_membrane, Colormap("bop_orange"))
merged_vs = (vs_nucleus_rgb + vs_membrane_rgb).clip(0, 1)

fluor_nucleus_rgb = render_rgb(fluor_nucleus, Colormap("green"))
fluor_membrane_rgb = render_rgb(fluor_membrane, Colormap("magenta"))
merged_fluor = (fluor_nucleus_rgb + fluor_membrane_rgb).clip(0, 1)

# Plot
# Show the individual channels and then fused in a grid
fig, ax = plt.subplots(2, 3, figsize=(15, 10))

# Virtual staining plots
ax[0, 0].imshow(vs_nucleus_rgb)
ax[0, 0].set_title("VS Nuclei")
ax[0, 1].imshow(vs_membrane_rgb)
ax[0, 1].set_title("VS Membrane")
ax[0, 2].imshow(merged_vs)
ax[0, 2].set_title("VS Nuclei+Membrane")

# Experimental fluorescence plots
ax[1, 0].imshow(fluor_nucleus_rgb)
ax[1, 0].set_title("Experimental Fluorescence Nuclei")
ax[1, 1].imshow(fluor_membrane_rgb)
ax[1, 1].set_title("Experimental Fluorescence Membrane")
ax[1, 2].imshow(merged_fluor)
ax[1, 2].set_title("Experimental Fluorescence Nuclei+Membrane")

# turnoff axis
for a in ax.flatten():
    a.axis("off")
plt.tight_layout()
plt.show()

# %% [markdown]
"""
## Responsible Use

We are committed to advancing the responsible development and use of artificial intelligence.
Please follow our [Acceptable Use Policy](https://virtualcellmodels.cziscience.com/acceptable-use-policy) when engaging with our services.

Should you have any security or privacy issues or questions related to the services,
please reach out to our team at [security@chanzuckerberg.com](mailto:security@chanzuckerberg.com) or [privacy@chanzuckerberg.com](mailto:privacy@chanzuckerberg.com) respectively.
"""



import numpy as np

from viscy.utils.image_utils import grid_sample_pixel_values, preprocess_image


def test_grid_sample_pixel_values():
    im = np.zeros((15, 20))
    row_ids, col_ids, sample_values = grid_sample_pixel_values(
        im,
        grid_spacing=5,
    )
    assert row_ids.tolist() == [5, 5, 5, 10, 10, 10]
    assert col_ids.tolist() == [5, 10, 15, 5, 10, 15]
    assert sample_values.tolist() == [0, 0, 0, 0, 0, 0]


def test_preprocess_image(self):
    im = np.zeros((5, 10, 15, 1))
    im[:, :5, :, :] = 10
    im_proc = preprocess_image(
        im,
        hist_clip_limits=(0, 100),
    )
    self.assertEqual(np.mean(im), np.mean(im_proc))
    self.assertTupleEqual(im_proc.shape, (5, 10, 15))


def test_preprocess_image_norm(self):
    im = np.zeros((5, 10, 15))
    im[:, :5, :] = 10
    im_proc = preprocess_image(
        im,
        normalize_im="dataset",
    )
    self.assertEqual(np.mean(im_proc), 0.0)
    self.assertTupleEqual(im.shape, im_proc.shape)


def test_preprocess_image_mask(self):
    im = np.zeros((5, 10, 15))
    im[:, :5, :] = 10
    im_proc = preprocess_image(
        im,
        is_mask=True,
    )
    self.assertEqual(np.mean(im_proc), 0.5)
    self.assertTupleEqual(im.shape, im_proc.shape)
    self.assertTrue(im_proc.dtype == bool)



import logging
import os
import sys
from datetime import datetime

import torch
from jsonargparse import lazy_instance
from lightning.pytorch import LightningDataModule, LightningModule
from lightning.pytorch.cli import LightningCLI
from lightning.pytorch.loggers import TensorBoardLogger

from viscy.trainer import VisCyTrainer


class VisCyCLI(LightningCLI):
    """Extending lightning CLI arguments and defualts."""

    @staticmethod
    def subcommands() -> dict[str, set[str]]:
        subcommands = LightningCLI.subcommands()
        subcommand_base_args = {"model"}
        subcommands["preprocess"] = subcommand_base_args
        subcommands["export"] = subcommand_base_args
        subcommands["precompute"] = subcommand_base_args
        return subcommands

    def add_arguments_to_parser(self, parser) -> None:
        parser.set_defaults(
            {
                "trainer.logger": lazy_instance(
                    TensorBoardLogger,
                    save_dir="",
                    version=datetime.now().strftime(r"%Y%m%d-%H%M%S"),
                    log_graph=True,
                )
            }
        )


def _setup_environment() -> None:
    """Set log level and TF32 precision."""
    log_level = os.getenv("VISCY_LOG_LEVEL", logging.INFO)
    logging.getLogger("lightning.pytorch").setLevel(log_level)
    torch.set_float32_matmul_precision("high")


def main() -> None:
    """
    Main Lightning CLI entry point.
    Parse log level and set TF32 precision.
    Set default random seed to 42.
    """
    _setup_environment()
    require_model = {"preprocess", "precompute"}.isdisjoint(sys.argv)
    require_data = {"preprocess", "precompute", "export"}.isdisjoint(sys.argv)
    _ = VisCyCLI(
        model_class=LightningModule,
        datamodule_class=LightningDataModule if require_data else None,
        trainer_class=VisCyTrainer,
        seed_everything_default=42,
        subclass_mode_model=require_model,
        subclass_mode_data=require_data,
        parser_kwargs={
            "description": "Computer vision models for single-cell phenotyping."
        },
    )


if __name__ == "__main__":
    main()



"""Test stage lightning module for comparing virtual staining and segmentations."""

import logging

from lightning.pytorch import LightningModule
from torchmetrics.functional import accuracy, jaccard_index
from torchmetrics.functional.segmentation import dice_score

from viscy.data.typing import SegmentationSample
from viscy.translation.evaluation_metrics import mean_average_precision

_logger = logging.getLogger("lightning.pytorch")


class SegmentationMetrics2D(LightningModule):
    """Test runner for 2D segmentation."""

    def __init__(self, aggregate_epoch: bool = False) -> None:
        super().__init__()
        self.aggregate_epoch = aggregate_epoch

    def test_step(self, batch: SegmentationSample, batch_idx: int) -> None:
        pred = batch["pred"]
        target = batch["target"]
        if not pred.shape[0] == 1 and target.shape[0] == 1:
            raise ValueError(
                f"Expected 2D segmentation, got {pred.shape[0]} and {target.shape[0]}"
            )
        pred = pred[0]
        target = target[0]
        pred_binary = pred > 0
        target_binary = target > 0
        coco_metrics = mean_average_precision(pred, target)
        _logger.debug(coco_metrics)
        self.logger.log_metrics(
            {
                "position": batch["position_idx"][0],
                "time": batch["time_idx"][0],
                "accuracy": (accuracy(pred_binary, target_binary, task="binary")),
                "dice": (
                    dice_score(
                        pred_binary.long()[None],
                        target_binary.long()[None],
                        num_classes=2,
                        input_format="index",
                    )
                ),
                "jaccard": (jaccard_index(pred_binary, target_binary, task="binary")),
                "mAP": coco_metrics["map"],
                "mAP_50": coco_metrics["map_50"],
                "mAP_75": coco_metrics["map_75"],
                "mAR_100": coco_metrics["mar_100"],
            }
        )



# %%

import lightning.pytorch as pl
from applications.infection_classification.classify_infection_2D import (
    SemanticSegUNet2D,
)

from viscy.data.hcs import HCSDataModule
from viscy.transforms import NormalizeSampled
from viscy.translation.predict_writer import HCSPredictionWriter

# %% # %% write the predictions to a zarr file

pred_datapath = "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/datasets/A549_63X/2024_02_04_A549_DENV_ZIKV_timelapse/0-train_test_data/2024_02_04_A549_DENV_ZIKV_timelapse_test_2D.zarr"

data_module = HCSDataModule(
    data_path=pred_datapath,
    source_channel=["RFP", "Phase3D"],
    target_channel=["Inf_mask"],
    split_ratio=0.7,
    z_window_size=1,
    architecture="2D",
    num_workers=1,
    batch_size=1,
    normalizations=[
        NormalizeSampled(
            keys=["RFP", "Phase3D"],
            level="fov_statistics",
            subtrahend="median",
            divisor="iqr",
        )
    ],
)

data_module.setup(stage="predict")

model = SemanticSegUNet2D(
    in_channels=2,
    out_channels=3,
    ckpt_path="/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/datasets/A549_63X/2024_02_04_A549_DENV_ZIKV_timelapse/1-model_train/logs/version_0/checkpoints/epoch=199-step=800.ckpt",
)

# %% perform prediction

output_path = "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/datasets/A549_63X/2024_02_04_A549_DENV_ZIKV_timelapse/2-predict_infection/2024_02_04_A549_DENV_ZIKV_timelapse_pred_2D_new.zarr"

trainer = pl.Trainer(
    default_root_dir="/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/datasets/A549_63X/2024_02_04_A549_DENV_ZIKV_timelapse/1-model_train/logs",
    callbacks=[HCSPredictionWriter(output_path, write_input=False)],
    devices=1,  # Set the number of GPUs to use. This avoids run-time exception from distributed training when the node has multiple GPUs
)

trainer.predict(
    model=model,
    datamodule=data_module,
    return_predictions=False,
)

# %%



import logging
from pathlib import Path
from typing import Literal

import torch
from iohub import open_ome_zarr
from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.utilities.compile import _maybe_unwrap_optimized
from torch.onnx import OperatorExportTypes

from viscy.preprocessing.precompute import precompute_array
from viscy.utils.meta_utils import generate_normalization_metadata

_logger = logging.getLogger("lightning.pytorch")


class VisCyTrainer(Trainer):
    def preprocess(
        self,
        data_path: Path,
        channel_names: list[str] | Literal[-1] = -1,
        num_workers: int = 1,
        block_size: int = 32,
        model: LightningModule | None = None,
    ):
        """
        Compute dataset statistics before training or testing for normalization.

        Parameters
        ----------
        data_path : Path
            Path to the HCS OME-Zarr dataset
        channel_names : list[str] | Literal[-1], optional
            Channel names to compute statistics for, by default -1
        num_workers : int, optional
            Number of CPU workers, by default 1
        block_size : int, optional
            Block size to subsample images, by default 32
        model: LightningModule, optional
            Ignored placeholder, by default None
        """
        if model is not None:
            _logger.warning("Ignoring model configuration during preprocessing.")
        with open_ome_zarr(data_path, layout="hcs", mode="r") as dataset:
            channel_indices = (
                [dataset.channel_names.index(c) for c in channel_names]
                if channel_names != -1
                else channel_names
            )
        generate_normalization_metadata(
            zarr_dir=data_path,
            num_workers=num_workers,
            channel_ids=channel_indices,
            grid_spacing=block_size,
        )

    def export(
        self,
        model: LightningModule,
        export_path: Path,
        ckpt_path: Path,
        format: str = "onnx",
    ):
        """
        Export the model for deployment (currently only ONNX is supported).

        Parameters
        ----------
        model : LightningModule
            Module to export.
        export_path : Path
            Output file name.
        ckpt_path : Path
            Model checkpoint path.
        format : str, optional
            Format (currently only ONNX is supported), by default "onnx".
        """
        if not format.lower() == "onnx":
            raise NotImplementedError(f"Export format '{format}'")
        model = _maybe_unwrap_optimized(model)
        self.strategy._lightning_module = model
        model.load_state_dict(torch.load(ckpt_path, weights_only=True)["state_dict"])
        model.eval()
        model.to_onnx(
            export_path,
            input_sample=model.example_input_array,
            export_params=True,
            opset_version=18,
            operator_export_type=OperatorExportTypes.ONNX_ATEN_FALLBACK,
            input_names=["input"],
            output_names=["output"],
            dynamic_axes={
                "input": {
                    0: "batch_size",
                    1: "channels",
                    3: "num_rows",
                    4: "num_cols",
                },
                "output": {
                    0: "batch_size",
                    1: "channels",
                    3: "num_rows",
                    4: "num_cols",
                },
            },
        )
        _logger.info(f"ONNX exported at {export_path}")

    def precompute(
        self,
        data_path: Path,
        output_path: Path,
        channel_names: list[str],
        subtrahends: list[Literal["mean"] | float],
        divisors: list[Literal["std"] | tuple[float, float]],
        image_array_key: str = "0",
        include_wells: list[str] | None = None,
        exclude_fovs: list[str] | None = None,
        model: LightningModule | None = None,
    ):
        precompute_array(
            data_path=data_path,
            output_path=output_path,
            channel_names=channel_names,
            subtrahends=subtrahends,
            divisors=divisors,
            image_array_key=image_array_key,
            include_wells=include_wells,
            exclude_fovs=exclude_fovs,
        )



import datetime
import os
import time

import torch

from viscy.utils.cli_utils import save_figure
from viscy.utils.normalize import hist_clipping


def log_feature(feature_map, name, log_save_folder, debug_mode):
    """
    If self.debug_mode, creates a visual of the given feature map, and saves it at
    'log_save_folder'
    If no log_save_folder specified, saves relative to working directory with timestamp.

    Currently only saving in working directory is supported.
    This is meant to be an analysis tool,
    and results should not be saved permanently.

    :param torch.tensor feature_map: feature map to create visualization log of
    :param str name: string
    :param str log_save_folder
    """
    try:
        if debug_mode:
            now = datetime.datetime.now()
            log_save_folder = (
                f"feature_map_{now.year}_{now.month}_{now.day}_{now.hour}_{now.minute}/"
            )
            logger = FeatureLogger(
                save_folder=log_save_folder,
                spatial_dims=3,
                grid_width=8,
            )
            logger.log_feature_map(
                feature_map,
                name,
                dim_names=["batch", "channels"],
            )
    except Exception:
        print(
            "Features of one input logged. Results saved at:"
            f"\n\t  {log_save_folder}. Will not log to avoid overwrite. \n"
        )


class FeatureLogger:
    def __init__(
        self,
        save_folder,
        spatial_dims=3,
        full_batch=False,
        save_as_grid=True,
        grid_width=0,
        normalize_by_grid=False,
    ):
        """
        Logger object for handling logging feature maps inside network architectures.

        Saves each 2d slice of a feature map in either a single grid per feature map
        stack or a directory tree of labeled slices.

        By default saves images into grid.

        :param str save_folder: output directory
        :param bool full_batch: if true, log all sample in batch (warning slow!),
                                defaults to False
        :param bool save_as_grid: if true feature maps are to be saved as a grid
                                containing all channels, else saved individually,
                                defaults to True
        :param int grid_width: desired width of grid if save_as_grid, by default
                                1/4 the number of channels, defaults to 0
        :param bool normalize_by_grid: if true, images saved in grid are normalized
                                to brightest pixel in entire grid, defaults to False

        """
        self.save_folder = save_folder
        self.spatial_dims = spatial_dims
        self.full_batch = full_batch
        self.save_as_grid = save_as_grid
        self.grid_width = grid_width
        self.normalize_by_grid = normalize_by_grid

        print("--- Initializing Logger ---")

    def log_feature_map(
        self,
        feature_map,
        feature_name,
        dim_names=[],
        vmax=0,
    ):
        """
        Creates a log of figures the given feature map tensor at 'save_folder'.
        Log is saved as images of feature maps in nested directory tree.

        By default _assumes that batch dimension is the first dimension_, and
        only logs the first sample in the batch, for performance reasons.

        Feature map logs cannot overwrite.

        :param torch.Tensor feature_map: feature map to log (typically 5d tensor)
        :parapm str feature_name: name of feature (will be used as dir name)
        :param list dim_names: names of each dimension, by default just numbers
        :param int spatial_dims: number of spatial dims, defaults to 3
        :param float vmax: maximum intensity to normalize figures by, by default
                        (if given 0) does relative normalization
        """
        # take tensor off of gpu and detach gradient
        feature_map = feature_map.detach().cpu()

        # handle dim names
        num_dims = len(feature_map.shape)
        if len(dim_names) == 0:
            dim_names = ["dim_" + str(i) for i in range(len(num_dims))]
        else:
            assert len(dim_names) + self.spatial_dims == num_dims, (
                "dim_names must be same length as nonspatial tensor dim length"
            )
        self.dim_names = dim_names

        # handle current feature_name
        feature_name = " " + feature_name if len(feature_name) > 0 else ""
        print(f"Logging{feature_name} feature map...", end="")
        self.feature_save_folder = os.path.join(self.save_folder, feature_name)

        start = time.time()
        self.map_feature_dims(feature_map, self.save_as_grid, vmax=vmax)

        print(f"done. Took {time.time() - start:.2f} seconds")

    def map_feature_dims(
        self,
        feature_map,
        save_as_grid,
        vmax=0,
        depth=0,
    ):
        """
        Recursive directory creation for organizing feature map logs

        If save_as_grid, will compile 'channels' (assumed to be last
        non-spatial dimension) into a single large image grid before saving.

        :param numpy.ndarray feature_map: see name
        :param str save_dir: see name
        :param bool save_as_grid: if true, saves images as channel grid
        :param float vmax: maximum intensity to normalize figures by
        :param int depth: recursion counter. depth in dimensions
        """

        for i in range(feature_map.shape[0]):
            if len(feature_map.shape) == 3:
                # individual saving
                z_slice = feature_map[i]
                save_figure(
                    z_slice.unsqueeze(0),
                    self.feature_save_folder,
                    f"z_slice_{i}",
                    vmax=vmax,
                )

            elif len(feature_map.shape) == 4 and save_as_grid:
                if feature_map.shape[0] == 1:
                    # if a single channel, can't save as grid
                    self.map_feature_dims(
                        feature_map,
                        save_as_grid=False,
                        depth=depth,
                    )
                else:
                    # grid saving
                    for z_depth in range(feature_map.shape[1]):
                        # set grid_width
                        if self.grid_width == 0:
                            if feature_map.shape[0] % 4 != 0:
                                raise AttributeError(
                                    f"number of channels ({feature_map.shape[0]}) "
                                    "must be divisible by 4 if grid_width unspecified"
                                )
                            self.grid_width = feature_map.shape[0] // 4
                        else:
                            if feature_map.shape[0] % self.grid_width != 0:
                                raise AttributeError(
                                    f"Grid width {self.grid_width} must be a divisor "
                                    f"of the number of channels {feature_map.shape[0]}"
                                )
                        # build grid by rows
                        # interleaving bars for ease of visualization
                        feature_map_grid = []
                        current_grid_row = []

                        for channel_num in range(feature_map.shape[0]):
                            # build rows by item in col
                            col_num = channel_num % self.grid_width
                            if col_num == 0 and channel_num != 0:
                                feature_map_grid.append(
                                    torch.cat(
                                        self.interleave_bars(current_grid_row, axis=1),
                                        dim=1,
                                    )
                                )
                                current_grid_row = []

                            # get 2d slice
                            map_slice = feature_map[channel_num, z_depth]

                            # norm slice to (0,1) unless normalize_by_grid
                            # which is done later
                            if not self.normalize_by_grid:
                                map_slice = torch.tensor(
                                    hist_clipping(
                                        map_slice.numpy(),
                                        min_percentile=0,
                                        max_percentile=100,
                                    )
                                )
                                map_slice = (
                                    map_slice - torch.min(map_slice)
                                ) / torch.max(map_slice)

                            current_grid_row.append(map_slice)
                        feature_map_grid.append(
                            torch.cat(
                                self.interleave_bars(current_grid_row, axis=1),
                                dim=1,
                            )
                        )
                        feature_map_grid = torch.cat(
                            self.interleave_bars(feature_map_grid, axis=0), dim=0
                        )
                        save_figure(
                            torch.unsqueeze(feature_map_grid, 0),
                            self.feature_save_folder,
                            f"z_slice_{z_depth}_channels_0-{feature_map.shape[0]}",
                            vmax=vmax,
                        )
                    break
            else:
                # tree recursion
                try:
                    name = os.path.join(
                        self.feature_save_folder, self.dim_names[depth] + f"_{i}"
                    )
                except Exception:
                    raise AttributeError("error in recursion")
                os.makedirs(name, exist_ok=False)
                self.map_feature_dims(
                    feature_map[i],
                    name,
                    save_as_grid,
                    depth=depth + 1,
                )

            if depth == 0 and not self.full_batch:
                break
        return

    def interleave_bars(self, arrays, axis, pixel_width=3, value=0):
        """
        Takes list of 2d torch tensors and interleaves bars to improve
        grid visualization quality.
        Assumes arrays are all of the same shape.

        :param list grid_arrays: list of tensors to place bars between
        :param int axis: axis on which to interleave bars (0 or 1)
        :param int pixel_width: width of bar, defaults to 3
        :param int value: value of bar pixels, defaults to 0
        """
        shape_match_axis = abs(axis - 1)
        length = arrays[0].shape[shape_match_axis]

        if axis == 0:
            bar = torch.ones((pixel_width, length)) * value
        elif axis == 1:
            bar = torch.ones((length, pixel_width)) * value
        else:
            raise AttributeError("axis must be 0 or 1")

        for i in range(1, len(arrays) * 2 - 1, 2):
            arrays.insert(i, bar)
        return arrays



from pathlib import Path

import pandas as pd
import torch
from lightning.pytorch import LightningModule
from lightning.pytorch.callbacks import BasePredictionWriter
from torch import nn
from torchmetrics.functional.classification import binary_accuracy, binary_f1_score

from viscy.representation.contrastive import ContrastiveEncoder
from viscy.utils.log_images import render_images


class ClassificationPredictionWriter(BasePredictionWriter):
    def __init__(self, output_path: Path):
        super().__init__("epoch")
        if Path(output_path).exists():
            raise FileExistsError(f"Output path {output_path} already exists.")
        self.output_path = output_path

    def write_on_epoch_end(self, trainer, pl_module, predictions, batch_indices):
        all_predictions = []
        for prediction in predictions:
            for key, value in prediction.items():
                if isinstance(value, torch.Tensor):
                    prediction[key] = value.detach().cpu().numpy().flatten()
            all_predictions.append(pd.DataFrame(prediction))
        pd.concat(all_predictions).to_csv(self.output_path, index=False)


class ClassificationModule(LightningModule):
    def __init__(
        self,
        encoder: ContrastiveEncoder,
        lr: float | None,
        loss: nn.Module | None = nn.BCEWithLogitsLoss(pos_weight=torch.tensor(1.0)),
    ):
        super().__init__()
        self.stem = encoder.stem
        self.backbone = encoder.encoder
        self.backbone.head.fc = nn.Linear(768, 1)
        self.loss = loss
        self.lr = lr
        self.example_input_array = torch.rand(2, 1, 15, 160, 160)

    def forward(self, x):
        x = self.stem(x)
        return self.backbone(x)

    def on_fit_start(self):
        self.train_examples = []
        self.val_examples = []

    def _fit_step(self, batch, stage: str, loss_on_step: bool):
        x, y = batch
        y_hat = self(x)
        loss = self.loss(y_hat, y)
        acc = binary_accuracy(y_hat, y)
        f1 = binary_f1_score(y_hat, y)
        self.log(f"loss/{stage}", loss, on_step=loss_on_step, on_epoch=True)
        self.log_dict(
            {f"metric/accuracy/{stage}": acc, f"metric/f1_score/{stage}": f1},
            on_step=False,
            on_epoch=True,
        )
        return loss, x[0, 0, x.shape[2] // 2].detach().cpu().numpy()

    def training_step(self, batch, batch_idx: int):
        loss, example = self._fit_step(batch, "train", loss_on_step=True)
        if batch_idx < 4:
            self.train_examples.append([example])
        return loss

    def validation_step(self, batch, batch_idx: int):
        loss, example = self._fit_step(batch, "val", loss_on_step=False)
        if batch_idx < 4:
            self.val_examples.append([example])
        return loss

    def predict_step(self, batch, batch_idx: int, dataloader_idx: int | None = None):
        x, y, indices = batch
        y_hat = nn.functional.sigmoid(self(x))
        indices["label"] = y
        indices["prediction"] = y_hat
        return indices

    def _log_images(self, examples, stage):
        image = render_images(examples)
        self.logger.experiment.add_image(
            f"{stage}/examples",
            image,
            global_step=self.current_epoch,
            dataformats="HWC",
        )

    def on_train_epoch_end(self):
        self._log_images(self.train_examples, "train")
        self.train_examples.clear()

    def on_validation_epoch_end(self):
        self._log_images(self.val_examples, "val")
        self.val_examples.clear()

    def configure_optimizers(self):
        return torch.optim.AdamW(self.parameters(), lr=self.lr)



# %% Imports
from pathlib import Path

from viscy.representation.embedding_writer import read_embedding_dataset
from viscy.representation.evaluation import load_annotation
from viscy.representation.evaluation.lca import fit_logistic_regression

# %%
TRAIN_FOVS = ["/A/3/7", "/A/3/8", "/A/3/9", "/B/4/6", "/B/4/7"]


model_embeddings = {
    "no-track": Path(
        "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_random_sampling2/feb_fixed_test_predict.zarr"
    ),
    "cell-aware-2ch": Path(
        "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_difcell_randomtime_sampling/Ver2_updateTracking_refineModel/predictions/Feb_2chan_128patch_32projDim/2chan_128patch_56ckpt_FebTest.zarr"
    ),
    "cell-aware-1ch": Path(
        "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_difcell_randomtime_sampling/Ver2_updateTracking_refineModel/predictions/Feb_1chan_128patch_32projDim/1chan_128patch_63ckpt_FebTest.zarr"
    ),
    "time-cell-aware": Path(
        "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/time_interval/predict/feb_test_time_interval_1_epoch_178.zarr"
    ),
}
path_annotations_infection = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/supervised_inf_pred/extracted_inf_state.csv"
)

# %%
for model_name, path_embedding in model_embeddings.items():
    print(f"Model: {model_name}")
    dataset = read_embedding_dataset(path_embedding)
    features = dataset["features"]

    infection = load_annotation(
        dataset,
        path_annotations_infection,
        "infection_state",
        {0.0: "background", 1.0: "uninfected", 2.0: "infected"},
    )

    log_reg = fit_logistic_regression(
        features,
        infection,
        train_fovs=TRAIN_FOVS,
        remove_background_class=True,
        scale_features=False,
        class_weight="balanced",
        solver="liblinear",
        random_state=42,
    )

# %%



"""Logging example images during training."""

from typing import Sequence

import numpy as np
from matplotlib.pyplot import get_cmap
from skimage.exposure import rescale_intensity
from torch import Tensor


def detach_sample(
    imgs: Sequence[Tensor], log_samples_per_batch: int
) -> list[list[np.ndarray]]:
    """Detach example images from the batch and convert them to numpy arrays.

    Parameters
    ----------
    imgs : Sequence[Tensor]
        Sequence of example images.
    log_samples_per_batch : int
        Number of first N samples in the sequence to detach.

    Returns
    -------
    list[list[np.ndarray]]
        Grid of example images.
        Rows are samples, columns are channels.
    """
    num_samples = min(imgs[0].shape[0], log_samples_per_batch)
    samples = []
    for i in range(num_samples):
        patches = []
        for img in imgs:
            patch = img[i].detach().cpu().numpy()
            patch = np.squeeze(patch[:, patch.shape[1] // 2])
            patches.append(patch)
        samples.append(patches)
    return samples


def render_images(
    imgs: Sequence[Sequence[np.ndarray]], cmaps: list[str] = []
) -> np.ndarray:
    """Render images in a grid.

    Parameters
    ----------
    imgs : Sequence[Sequence[np.ndarray]]
        Grid of images to render, output of `detach_sample`.
    cmaps : list[str], optional
        Colormaps for each column, by default []

    Returns
    -------
    np.ndarray
        Rendered RGB images grid.
    """
    images_grid = []
    for sample_images in imgs:
        images_row = []
        for i, image in enumerate(sample_images):
            if cmaps:
                cm_name = cmaps[i]
            else:
                cm_name = "gray" if i == 0 else "inferno"
            if image.ndim == 2:
                image = image[np.newaxis]
            for channel in image:
                channel = rescale_intensity(channel, out_range=(0, 1))
                render = get_cmap(cm_name)(channel, bytes=True)[..., :3]
                images_row.append(render)
        images_grid.append(np.concatenate(images_row, axis=1))
    return np.concatenate(images_grid, axis=0)



import bisect
import logging
from collections import defaultdict
from enum import Enum
from typing import Literal, Sequence

import torch
from lightning.pytorch import LightningDataModule
from lightning.pytorch.utilities.combined_loader import CombinedLoader
from monai.data import ThreadDataLoader
from torch.utils.data import ConcatDataset, DataLoader, Dataset

from viscy.data.distributed import ShardedDistributedSampler
from viscy.data.hcs import _collate_samples

_logger = logging.getLogger("lightning.pytorch")


class CombineMode(Enum):
    MIN_SIZE = "min_size"
    MAX_SIZE_CYCLE = "max_size_cycle"
    MAX_SIZE = "max_size"
    SEQUENTIAL = "sequential"


class CombinedDataModule(LightningDataModule):
    """Wrapper for combining multiple data modules.
    For supported modes, see ``lightning.pytorch.utilities.combined_loader``.

    Parameters
    ----------
    data_modules : Sequence[LightningDataModule]
        data modules to combine
    train_mode : CombineMode, optional
        mode in training stage, by default CombineMode.MAX_SIZE_CYCLE
    val_mode : CombineMode, optional
        mode in validation stage, by default CombineMode.SEQUENTIAL
    test_mode : CombineMode, optional
        mode in testing stage, by default CombineMode.SEQUENTIAL
    predict_mode : CombineMode, optional
        mode in prediction stage, by default CombineMode.SEQUENTIAL
    """

    def __init__(
        self,
        data_modules: Sequence[LightningDataModule],
        train_mode: CombineMode = CombineMode.MAX_SIZE_CYCLE,
        val_mode: CombineMode = CombineMode.SEQUENTIAL,
        test_mode: CombineMode = CombineMode.SEQUENTIAL,
        predict_mode: CombineMode = CombineMode.SEQUENTIAL,
    ):
        super().__init__()
        self.data_modules = data_modules
        self.train_mode = CombineMode(train_mode).value
        self.val_mode = CombineMode(val_mode).value
        self.test_mode = CombineMode(test_mode).value
        self.predict_mode = CombineMode(predict_mode).value
        self.prepare_data_per_node = True

    def prepare_data(self):
        for dm in self.data_modules:
            dm.trainer = self.trainer
            dm.prepare_data()

    def setup(self, stage: Literal["fit", "validate", "test", "predict"]):
        for dm in self.data_modules:
            dm.setup(stage)

    def train_dataloader(self):
        return CombinedLoader(
            [dm.train_dataloader() for dm in self.data_modules], mode=self.train_mode
        )

    def val_dataloader(self):
        return CombinedLoader(
            [dm.val_dataloader() for dm in self.data_modules], mode=self.val_mode
        )

    def test_dataloader(self):
        return CombinedLoader(
            [dm.test_dataloader() for dm in self.data_modules], mode=self.test_mode
        )

    def predict_dataloader(self):
        return CombinedLoader(
            [dm.predict_dataloader() for dm in self.data_modules],
            mode=self.predict_mode,
        )


class BatchedConcatDataset(ConcatDataset):
    def __getitem__(self, idx):
        raise NotImplementedError

    def _get_sample_indices(self, idx: int) -> tuple[int, int]:
        if idx < 0:
            if -idx > len(self):
                raise ValueError(
                    "absolute value of index should not exceed dataset length"
                )
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        return dataset_idx, sample_idx

    def __getitems__(self, indices: list[int]) -> list:
        grouped_indices = defaultdict(list)
        for idx in indices:
            dataset_idx, sample_indices = self._get_sample_indices(idx)
            grouped_indices[dataset_idx].append(sample_indices)
        _logger.debug(f"Grouped indices: {grouped_indices}")
        sub_batches = []
        for dataset_idx, sample_indices in grouped_indices.items():
            sub_batch = self.datasets[dataset_idx].__getitems__(sample_indices)
            sub_batches.extend(sub_batch)
        return sub_batches


class ConcatDataModule(LightningDataModule):
    """
    Concatenate multiple data modules.

    The concatenated data module will have the same batch size and number of workers
    as the first data module. Each element will be sampled uniformly regardless of
    their original data module.

    Parameters
    ----------
    data_modules : Sequence[LightningDataModule]
        Data modules to concatenate.
    """

    _ConcatDataset = ConcatDataset

    def __init__(self, data_modules: Sequence[LightningDataModule]):
        super().__init__()
        self.data_modules = data_modules
        self.num_workers = data_modules[0].num_workers
        self.batch_size = data_modules[0].batch_size
        self.persistent_workers = data_modules[0].persistent_workers
        self.prefetch_factor = data_modules[0].prefetch_factor
        self.pin_memory = data_modules[0].pin_memory
        for dm in data_modules:
            if dm.num_workers != self.num_workers:
                raise ValueError("Inconsistent number of workers")
            if dm.batch_size != self.batch_size:
                raise ValueError("Inconsistent batch size")
        self.prepare_data_per_node = True

    def prepare_data(self):
        for dm in self.data_modules:
            dm.trainer = self.trainer
            dm.prepare_data()

    def setup(self, stage: Literal["fit", "validate", "test", "predict"]):
        self.train_patches_per_stack = 0
        for dm in self.data_modules:
            dm.setup(stage)
            if patches := getattr(dm, "train_patches_per_stack", 0):
                if self.train_patches_per_stack == 0:
                    self.train_patches_per_stack = patches
                elif self.train_patches_per_stack != patches:
                    raise ValueError("Inconsistent patches per stack")
        if stage != "fit":
            raise NotImplementedError("Only fit stage is supported")
        self.train_dataset = self._ConcatDataset(
            [dm.train_dataset for dm in self.data_modules]
        )
        self.val_dataset = self._ConcatDataset(
            [dm.val_dataset for dm in self.data_modules]
        )

    def _dataloader_kwargs(self) -> dict:
        return {
            "num_workers": self.num_workers,
            "persistent_workers": self.persistent_workers,
            "prefetch_factor": self.prefetch_factor if self.num_workers else None,
            "pin_memory": self.pin_memory,
        }

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset,
            shuffle=True,
            batch_size=self.batch_size // self.train_patches_per_stack,
            collate_fn=_collate_samples,
            drop_last=True,
            **self._dataloader_kwargs(),
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_dataset,
            shuffle=False,
            batch_size=self.batch_size,
            drop_last=False,
            **self._dataloader_kwargs(),
        )


class BatchedConcatDataModule(ConcatDataModule):
    _ConcatDataset = BatchedConcatDataset

    def train_dataloader(self):
        return ThreadDataLoader(
            self.train_dataset,
            use_thread_workers=True,
            batch_size=self.batch_size,
            shuffle=True,
            drop_last=True,
            **self._dataloader_kwargs(),
        )

    def val_dataloader(self):
        return ThreadDataLoader(
            self.val_dataset,
            use_thread_workers=True,
            batch_size=self.batch_size,
            shuffle=False,
            drop_last=False,
            **self._dataloader_kwargs(),
        )


class CachedConcatDataModule(LightningDataModule):
    def __init__(self, data_modules: Sequence[LightningDataModule]):
        super().__init__()
        self.data_modules = data_modules
        self.num_workers = data_modules[0].num_workers
        self.batch_size = data_modules[0].batch_size
        for dm in data_modules:
            if dm.num_workers != self.num_workers:
                raise ValueError("Inconsistent number of workers")
            if dm.batch_size != self.batch_size:
                raise ValueError("Inconsistent batch size")
        self.prepare_data_per_node = True

    def prepare_data(self):
        for dm in self.data_modules:
            dm.trainer = self.trainer
            dm.prepare_data()

    def setup(self, stage: Literal["fit", "validate", "test", "predict"]):
        self.train_patches_per_stack = 0
        for dm in self.data_modules:
            dm.setup(stage)
            if patches := getattr(dm, "train_patches_per_stack", 1):
                if self.train_patches_per_stack == 0:
                    self.train_patches_per_stack = patches
                elif self.train_patches_per_stack != patches:
                    raise ValueError("Inconsistent patches per stack")
        if stage != "fit":
            raise NotImplementedError("Only fit stage is supported")
        self.train_dataset = ConcatDataset(
            [dm.train_dataset for dm in self.data_modules]
        )
        self.val_dataset = ConcatDataset([dm.val_dataset for dm in self.data_modules])

    def _maybe_sampler(
        self, dataset: Dataset, shuffle: bool
    ) -> ShardedDistributedSampler | None:
        return (
            ShardedDistributedSampler(dataset, shuffle=shuffle)
            if torch.distributed.is_initialized()
            else None
        )

    def train_dataloader(self) -> DataLoader:
        sampler = self._maybe_sampler(self.train_dataset, shuffle=True)
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=False if sampler else True,
            sampler=sampler,
            persistent_workers=True if self.num_workers > 0 else False,
            num_workers=self.num_workers,
            drop_last=True,
            collate_fn=lambda x: x,
        )

    def val_dataloader(self) -> DataLoader:
        sampler = self._maybe_sampler(self.val_dataset, shuffle=False)
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            sampler=sampler,
            persistent_workers=True if self.num_workers > 0 else False,
            num_workers=self.num_workers,
            drop_last=False,
            collate_fn=lambda x: x,
        )
